@@ -4,13 +4,13 @@ use Carp ();
 ############################################################################
 package Net::OpenID::ClaimedIdentity;
 BEGIN {
-  $Net::OpenID::ClaimedIdentity::VERSION = '1.100099_001';
+  $Net::OpenID::ClaimedIdentity::VERSION = '1.100099_002';
 }
 use fields (
     'identity',         # the canonical URL that was found, following redirects
-    'server',           # author-identity identity server endpoint
+    'server',           # author-identity identity provider endpoint
     'consumer',         # ref up to the Net::OpenID::Consumer which generated us
-    'delegate',         # the delegated URL actually asserted by the server
+    'delegate',         # the delegated URL actually asserted by the provider
     'protocol_version', # The version of the OpenID Authentication Protocol that is used
     'semantic_info',    # Stuff that we've discovered in the identifier page's metadata
     'extension_args',   # Extension arguments that the caller wants to add to the request
@@ -213,7 +213,7 @@ Net::OpenID::ClaimedIdentity - A not-yet-verified OpenID identity
 
 =head1 VERSION
 
-version 1.100099_001
+version 1.100099_002
 
 =head1 SYNOPSIS
 
@@ -258,13 +258,13 @@ check_url, though.
 
 =item $id_server = $cident->B<identity_server>
 
-Returns the identity server that will assert whether or not this
+Returns the identity provider that will assert whether or not this
 claimed identity is valid, and sign a message saying so.
 
 =item $url = $cident->B<delegated_url>
 
 If the claimed URL is using delegation, this returns the delegated identity that will
-actually be sent to the identity server.
+actually be sent to the identity provider.
 
 =item $version = $cident->B<protocol_version>
 
@@ -302,10 +302,11 @@ validate their identity.  The options to put in %opts are:
 
 =item C<return_to>
 
-The URL that the identity server should redirect the user with either
-a verified identity signature -or- a user_setup_url (if the assertion
-couldn't be made).  This URL may contain query parameters, and the
-identity server must preserve them.
+The URL that the identity provider should redirect the user with
+either a verified identity signature -or- a setup_needed message
+(indicating actual interaction with the user is required before an
+assertion can be made).  This URL may contain query parameters, and
+the identity provider must preserve them.
 
 =item C<trust_root>
 
@@ -319,18 +320,19 @@ beginning of the host, like C<http://*.example.com/>
 =item C<delayed_return>
 
 If set to a true value, the check_url returned will indicate to the
-user's identity server that it has permission to control the user's
+user's identity provider that it has permission to control the user's
 user-agent for awhile, giving them real pages (not just redirects) and
-lets them bounce around the identity server site for awhile until
+lets them bounce around the identity provider site for awhile until
 the requested assertion can be made, and they can finally be redirected
 back to your return_to URL above.
 
-The default value, false, means that the identity server will
+The default value, false, means that the identity provider will
 immediately return to your return_to URL with either a "yes" or "no"
 answer.  In the "no" case, you'll instead have control of what to do,
-and you'll be sent the identity server's user_setup_url where you'll
-have to somehow send the user (be it link, redirect, or pop-up
-window).
+whether to retry the request with C<delayed_return> set true 
+(the only way to proceed in version 2.0) or to somehow send 
+(be it link, redirect, or pop-up window) the user the provider's
+user_setup_url (which is made available in version 1.0/1.1).
 
 When writing a dynamic "AJAX"-style application, you can't use
 delayed_return because the remote site can't usefully take control of
